#!/usr/bin/env python
#
# __init__.py - The pyfeeds package.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

__version__ = '0.9.5'
"""The pyfeeds version number. """


from .common import (
    fslbin,
    readfsf,
    writefsf,
)


from .evaluate import (
    evalImage,
    evalNumericalText,
    evalVectorImage,
    evalMD5
)



#!/usr/bin/env python
#
# test_evaluate.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os
import os.path as op

from . import tempdir, makepaths, maketest, makepyfeeds, CaptureStdout

from pyfeeds import testing, evaluate


def test_evaluateTestAgainstBenchmark():
    with tempdir():
        maketest(  'tests/test1/feedsRun',   outputs=['file1', 'file2'])
        maketest(  'tests/test2/feedsRun.1', outputs=['file1', 'file2'])
        maketest(  'tests/test2/feedsRun.2', outputs=['file1', 'file2'])
        os.chmod(  'tests/test1/feedsRun'  , 0o755)
        os.chmod(  'tests/test2/feedsRun.1', 0o755)
        os.chmod(  'tests/test2/feedsRun.2', 0o755)
        os.mkdir('output/')
        os.mkdir('benchmarks/')

        makepaths(['benchmarks/test1/file1',
                   'benchmarks/test1/file2',
                   'benchmarks/test2/1/file1',
                   'benchmarks/test2/1/file2',
                   'benchmarks/test2/2/file1',
                   'benchmarks/test2/2/file2'])

        pyf = makepyfeeds(command='run',
                          testDir=['tests'],
                          outputDir='output',
                          benchmarkDir='benchmarks',
                          overwrite=True)

        for test in pyf.tests:
            assert testing.runTest(pyf,
                                   test,
                                   'output',
                                   None,
                                   'benchmarks',
                                   os.environ.copy(),
                                   False) == (True, 0)

        for resfile in ['output/test1/feedsResults.log',
                        'output/test2/1/feedsResults.log',
                        'output/test2/2/feedsResults.log']:
            assert op.exists(resfile)
            with open(resfile) as f:
                lines = f.readlines()
            lines = [l.strip() for l in lines]
            lines = [l for l in lines if l != '']
            assert len(lines) == 3
            for l in lines[1:]:
                assert 'PASS' in l



#!/usr/bin/env python
#
# __init__.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os
import sys
import os.path as op
import contextlib
import tempfile

from io import StringIO

from pyfeeds import main

import argparse


class CaptureStdout(object):
    """Context manager which captures stdout and stderr. """

    def __init__(self):
        self.reset()

    def reset(self):
        self.__mock_stdout = StringIO('')
        self.__mock_stderr = StringIO('')

    def __enter__(self):
        self.__real_stdout = sys.stdout
        self.__real_stderr = sys.stderr

        sys.stdout = self.__mock_stdout
        sys.stderr = self.__mock_stderr


    def __exit__(self, *args, **kwargs):
        sys.stdout = self.__real_stdout
        sys.stderr = self.__real_stderr

        if args[0] is not None:
            print('Error')
            print('stdout:')
            print(self.stdout)
            print('stderr:')
            print(self.stderr)

        return False

    @property
    def stdout(self):
        self.__mock_stdout.seek(0)
        return self.__mock_stdout.read()

    @property
    def stderr(self):
        self.__mock_stderr.seek(0)
        return self.__mock_stderr.read()


@contextlib.contextmanager
def tempdir():
    prevdir = os.getcwd()
    with tempfile.TemporaryDirectory() as td:
        try:
            os.chdir(td)
            yield td
        finally:
            os.chdir(prevdir)


def makepaths(paths):
    for path in paths:
        dirname = op.dirname(path)
        if dirname != '':
            os.makedirs(dirname, exist_ok=True)
        with open(path, 'w'):
            pass


def maketest(filename, returnCode=0, inputs=None, outputs=None, stdout=None):

    if outputs is None:
        outputs = []
    if stdout is None:
        stdout = []

    dirname = op.dirname(filename)

    os.makedirs(dirname, exist_ok=True)

    with open(filename, 'wt') as f:
        f.write('#!/bin/bash\n')
        f.write('outdir=$1\n')

        for output in outputs:
            f.write('mkdir -p $(dirname $outdir/{})\n'.format(output))
            f.write('touch $outdir/{}\n'.format(output))

        for line in stdout:
            f.write('echo "{}"\n'.format(line))

        f.write('exit {}\n'.format(returnCode))


    if inputs is not None:
        with open(op.join(dirname, 'feedsInputs'), 'wt') as f:
            for i in inputs:
                f.write(i + '\n')


def makepyfeeds(**kwargs):
    args = argparse.Namespace(**kwargs)
    cfg  = argparse.Namespace()
    return main.Pyfeeds(args, cfg)



#!/usr/bin/env python
#
# main.py - The FSL Evaluation and Example Data Suite, now in Python!
#
# Authors: Matthew Webster and Paul McCarthy, FMRIB.
#
"""The FSL Evaluation and Example Data Suite (FEEDS), now in Python!

Pyfeeds (the FMRIB Evaluation and Example Data Suite) is a framework for
running and managing tests for the FSL code base.


See the following documents  for details on pyfeeds:

  - Using pyfeeds:                  ``doc/using_pyfeeds.md``
  - How to write a pyfeeds test:    ``doc/writing_a_test.md``
  - How pyfeeds works:              ``doc/how_pyfeeds_works.md``
  - The pyfeeds configuration file: ``doc/configuring_pyfeeds.md``


-----------------
Pyfeeds internals
-----------------


The pyfeeds logic is split across a few separate modules:

 - This module contains the :class:`Pyfeeds` class, the :func:`main` function
   (the pyfeeds entry point), and argument parsing logic.

 - The :mod:`.testing` module contains the :class:`.Test` class, and
   functions for identifying and running tests.

 - The :mod:`.evaluate` module contains functions for evaluating test results.

 - The :mod:`.hashing` module contains functions for test verification (to
   identify when tests or test data have changed).

 - The :mod:`.common` module contains other miscellaneous functions.


The :func:`main` function is the entry point for pyfeeds. It parses arguments
and a configuration file (if one was specified), and creates a
:class:`Pyfeeds` object. This object combines the command line and
configuration file settings, identifies all of the tests in the provided test
directories, and creates a :class:`.Test` object for each of them. The
:func:`main` function then calls one of the following functions, depending
upon which sub-command the user has specified:


  ============= =============================
  *Command*     *Function*
  ``list``      :func:`.testing.listTests`
  ``run``       :func:`.testing.runTests`
  ``bundle``    :func:`.testing.bundleTests`
  ``genhash``   :func:`.hashing.genHashes`
  ``checkhash`` :func:`.hashing.checkHashes`
  ``compare``   :func:`.evaluate.compareDirs`
  ============= =============================
"""


from __future__ import print_function
from __future__ import division

import os.path as op
import            os
import            sys
import            logging
import            fnmatch
import            argparse
import            datetime
import            functools
import            configparser
from collections import OrderedDict

from . import __version__ as pyfeeds_version
from . import common
from . import testing
from . import hashing
from . import evaluate
from . import imagecache


log = logging.getLogger()


ARG_SEP = ':'
"""String used as a separator character in the `--selfEval`, `--evalRoutines`,
`--exclude`, `--tolerances`, and `--fileGroups` command-line arguments.
"""


class Pyfeeds(object):
    """The ``Pyfeeds`` class is used as a container for:

        - command line arguments
        - settings read from the pyfeeds configuration file
        - the list of tests to be considered


    A ``Pyfeeds`` instance is created in the :func:`main` function, and passed
    around the various parts of ``pyfeeds``. Only one ``Pyfeeds`` instance
    should ever exist.
    """


    def __init__(self, args, cfg):
        """Create the ``Pyfeeds`` instance.

        Combines the command line and configuration file arguments, and
        creates a :class:`.Test` instance for all identified tests.

        :arg args: ``argparse.Namespace`` object containing parsed command
                   line arguments. It is assumed that, if an argument was
                   not passed in on the command line, it is set to ``None``.

        :arg cfg: ``argparse.Namespace`` object containing settings loaded
                   from a pyfeeds configuration file.  It is assumed that, if
                   an argument was not present in the configuration file, it
                   is present in this object.
        """

        # Default values for all settings
        self.verbose          = False
        self.quiet            = False
        self.includeTests     = None
        self.excludeTests     = None
        self.testDir          = []
        self.inputDir         = None
        self.benchmarkDir     = None
        self.outputDir        = None
        self.overwrite        = False
        self.outputFile       = None
        self.config           = None
        self.hashFile         = None
        self.jobs             = 1
        self.updateHashes     = False
        self.skipHashes       = False
        self.forceHashes      = False
        self.sandboxDir       = None
        self.leaveSandboxes   = False
        self.selfEval         = []
        self.evalRoutines     = {'*' : ['evalMD5']}
        self.fileGroups       = []
        self.exclude          = []
        self.tolerances       = {}
        self.defaultTolerance = 0.05
        self.genHashes        = False
        self.cacheSize        = 32768
        self.imageCache       = None

        for key, val in cfg.__dict__.items():

            log.debug('Applying config file setting:  {} = {}'.format(key,
                                                                      val))
            setattr(self, key, val)

        for key, val in args.__dict__.items():
            if val is None:
                continue
            if key == 'testDir' and len(val) == 0:
                continue

            log.debug('Applying command line setting: {} = {}'.format(key,
                                                                      val))

            setattr(self, key, val)

        # If no test dirs were passed,
        # set args.testDir to cwd
        if self.command != 'compare' and \
           (not hasattr(self, 'testDir') or len(self.testDir) == 0):
            self.testDir = [os.getcwd()]

        # Hashfile defaults to inputDir/.feedsHashes,
        # but we record whether the user actually
        # specified a hash file - if they don't
        # explicitly specify a hashfile, the genhash
        # command will print to stdout.
        self.hashFileProvided = self.hashFile is not None

        if self.hashFile is None and self.inputDir is not None:
            hashFile = op.join(self.inputDir, '.feedsHashes')
            if op.exists(hashFile):
                self.hashFile = hashFile

        # Figure out when the hash file was last generated
        if self.hashFile is not None and op.exists(self.hashFile):
            mtime = op.getmtime(self.hashFile)
            mtime = datetime.datetime.fromtimestamp(mtime)
            self.hashesLastGenerated = mtime
        else:
            self.hashesLastGenerated = None

        # Add a "*/" at the beginning of every file
        # name pattern, because the evaluate module
        # tests absolute file paths.
        def absify(p):
            return '*{}{}'.format(op.sep, p)

        self.exclude      = [absify(e) for e in self.exclude]
        self.evalRoutines = {absify(k) : v
                             for k, v in self.evalRoutines.items()}
        self.tolerances   = {absify(k) : v
                             for k, v in self.tolerances.items()}

        # All files in a group are currently
        # assumed to be in the same directory
        self.fileGroups = [[op.basename(p) for p in g]
                           for g in self.fileGroups]

        self.tests = self.__findTests(self.testDir)
        self.__validate()

        self.imageCache = imagecache.ImageCache(maxsize=self.cacheSize)


    def __findTests(self, testDirs):
        """Finds all pyfeeds tests in the given set of directories.

        :returns: A list of :class:`.Test` instances.
        """

        # Find all test sub-directories contained within
        # the directories specified by the user
        testDirs = [testing.findTestDirs(d) for d in testDirs]

        if len(testDirs) == 0:
            return []

        testDirs = functools.reduce(lambda a, b: a + b, testDirs)
        testDirs = sorted(set(testDirs))

        if len(testDirs) == 0:
            return []

        # remove any tests that don't
        # match the include patterns
        if self.includeTests is not None:
            filtered = []
            for td in testDirs:
                if any([fnmatch.fnmatch(td, p) for p in self.includeTests]):
                    filtered.append(td)
            testDirs = filtered

        # Remove any tests which match
        # the exclude patterns
        if self.excludeTests is not None:
            filtered = []
            for td in testDirs:
                if not any([fnmatch.fnmatch(td, p)
                            for p in self.excludeTests]):
                    filtered.append(td)
            testDirs = filtered

        if len(testDirs) == 0:
            return []

        # Find the deepest directory in
        # the file system which contains
        # all of the detected tests.
        rootDir = common.findCommonAncestor(testDirs)
        log.debug('Root of all test directories: {}'.format(rootDir))

        # Create a Test object for each
        # one of the test directories
        tests = []
        for td in testDirs:
            for script in testing.findTestScripts(td):
                try:
                    tests.append(testing.Test(script, rootDir))
                except Exception as e:
                    log.warning('Error loading test {} ({})'.format(td, e))

        return tests


    def __validate(self):
        """Performs some simple validation to make sure that the user
        has specified all required arguments.
        """

        notests      = 'No tests specified!'
        noinput      = 'Input data directory not specified or invalid!'
        nooutputdir  = 'Output directory not specified!'
        outputexist  = 'Output directory already exists!'
        nobenchmark  = 'Benchmark data directory not specified or invalid!'
        badroutine   = '{} is not a valid evaluation routine!'

        command      = self.command
        tests        = self.tests
        inputDir     = self.inputDir
        benchmarkDir = self.benchmarkDir
        outputDir    = self.outputDir
        ovw          = self.overwrite

        # all commands except for
        # compare need some tests
        if command != 'compare' and len(tests) == 0:
            raise RuntimeError(notests)

        if command in ('run', 'bundle'):

            if outputDir is None:
                raise RuntimeError(nooutputdir)
            if op.exists(outputDir) and not ovw:
                raise RuntimeError(outputexist)
            if (inputDir is not None) and (not op.exists(inputDir)):
                raise RuntimeError(noinput)
            if (benchmarkDir is not None) and (not op.exists(benchmarkDir)):
                raise RuntimeError(nobenchmark)

        elif command == 'compare':

            if inputDir     is None:        raise RuntimeError(noinput)
            if benchmarkDir is None:        raise RuntimeError(nobenchmark)
            if not op.exists(inputDir):     raise RuntimeError(noinput)
            if not op.exists(benchmarkDir): raise RuntimeError(nobenchmark)

        elif command in ('genhash', 'checkhash'):
            if (inputDir is not None) and (not op.exists(inputDir)):
                raise RuntimeError(noinput)

        for testName, routines in self.evalRoutines.items():
            for routine in routines:
                if not hasattr(evaluate, routine):
                    raise RuntimeError(badroutine.format(routine))


def main(argv=None):
    """Pyfeeds entry point. This function does the following:

      1. Parses command line arguments (via the :func:`parseArgs` function).

      2. Loads a configuration file if one was specified (via the
         :func:`loadPyfeedsConfig` function).

      3. Creates a :class:`Pyfeeds` object.

      4. Calls the command function which corresponds to the sub-command
         that was specified by the user:

         .. autosummary::
            :nosignatures:

            .testing.listTests
            .testing.runTests
            .testing.bundleTests
            .hashing.genHashes
            .hashing.checkHashes
            .evaluate.compareDirs

    :arg argv: Command line arguments. If not provided, ``sys.argv`` is used.
    """

    args = parseArgs(argv)
    cfg  = loadPyfeedsConfig(args.config)

    configLogging(args, cfg)

    pyfeeds = Pyfeeds(args, cfg)

    log.info('pyfeeds %s', pyfeeds_version)

    # Do that thing we were asked to do
    if   args.command == 'list':      testing .listTests(  pyfeeds)
    elif args.command == 'run':       testing .runTests(   pyfeeds)
    elif args.command == 'bundle':    testing .bundleTests(pyfeeds)
    elif args.command == 'genhash':   hashing .genHashes(  pyfeeds)
    elif args.command == 'checkhash': hashing .checkHashes(pyfeeds)
    elif args.command == 'compare':   evaluate.compareDirs(pyfeeds)


def configLogging(args, cfg):
    """Configures logging, including verbosity, and streaming to a file.
    """
    def getarg(name):

        argval = getattr(args, name, None)
        cfgval = getattr(cfg,  name, None)

        if argval: return argval
        else:      return cfgval

    verbose = getarg('verbose')
    quiet   = getarg('quiet')
    logFile = getarg('logFile')

    logger = logging.getLogger()
    fmt    = logging.Formatter('%(levelname)8.8s '
                               '%(filename)20.20s '
                               '%(lineno)4d: '
                               '%(funcName)-15.15s - '
                               '%(message)s')

    if logFile is None: handler = logging.StreamHandler()
    else:               handler = logging.FileHandler(logFile)

    handler.setFormatter(fmt)
    logger.addHandler(handler)

    if   verbose: logger.setLevel(logging.DEBUG)
    elif quiet:   logger.setLevel(logging.CRITICAL + 1)
    else:         logger.setLevel(logging.INFO)


def parseArgs(argv):
    """Parses command line arguments.

    :arg argv:  List of arguments - if None, sys.argv is used.

    :returns:   An ``argparse.Namespace`` object.
    """

    if argv is None:
        argv = sys.argv[1:]

    # If the user gave us nothing,
    # give them some help
    if len(argv) == 0:
        argv = ['--help']

    usages = {
        'mainParser' :
        'pyfeeds command            [options] [testDir ...]\n'
        'pyfeeds command -c cfgfile [options] [testDir ...]\n',

        'list' :
        'pyfeeds list [testDir ...]',

        'run' :
        'pyfeeds run\n'
        '          -i inputDir\n'
        '          -b benchmarkDIr\n'
        '          -o outputDir\n'
        '          [options] [testDir ...]',

        'bundle' :
        'pyfeeds bundle\n'
        '          -i inputDir\n'
        '          -b benchmarkDir\n'
        '          -o outputDir\n'
        '          [options] [testDir ...]',

        'genhash' :
        'pyfeeds genhash -i inputDir [options] [testDir ...]',

        'checkhash' :
        'pyfeeds checkhash -i inputDir [options] [testDir ...]',

        'compare' :
        'pyfeeds compare [options] inputDir benchmarkDir\n',
    }

    flags = {

        # global options
        'version'        : ('-V',  '--version'),
        'verbose'        : ('-v',  '--verbose'),
        'quiet'          : ('-q',  '--quiet'),
        'logFile'        : ('-lf', '--logFile'),
        'includeTests'   : ('-it', '--includeTests'),
        'excludeTests'   : ('-et', '--excludeTests'),
        'cacheSize'      : ('-cs', '--cacheSize'),
        'testDir'        : 'testDir',

        # shared options
        'inputDir'       : ('-i',  '--inputDir'),
        'benchmarkDir'   : ('-b',  '--benchmarkDir'),
        'outputDir'      : ('-o',  '--outputDir'),
        'overwrite'      : ('-w',  '--overwrite'),
        'config'         : ('-c',  '--config'),
        'hashFile'       : ('-a',  '--hashFile'),
        'forceHashes'    : ('-fh', '--forceHashes'),

        # run options
        'jobs'             : ('-j', '--jobs'),
        'updateHashes'     : ('-u', '--updateHashes'),
        'skipHashes'       : ('-k', '--skipHashes'),
        'sandboxDir'       : ('-n', '--sandboxDir'),
        'leaveSandboxes'   : ('-l', '--leaveSandboxes'),
        'selfEval'         : ('-s', '--selfEval'),
        'evalRoutines'     : ('-e', '--evalRoutines'),
        'exclude'          : ('-x', '--exclude'),
        'tolerances'       : ('-t', '--tolerances'),
        'defaultTolerance' : ('-d', '--defaultTolerance'),
        'fileGroups'       : ('-f', '--fileGroups'),

        # bundle options
        'genHashes'        : ('-g',  '--genHashes'),

        # compare options
        'outputFile'       : ('-o', '--outputFile'),
    }

    helps = {
        # Commands
        'list'             : 'List available tests',
        'run'              : 'Run tests',
        'bundle'           : 'Bundle tests',
        'genhash'          : 'Generate test hashes',
        'checkhash'        : 'Validate test hashes',
        'compare'          : 'Compare two directories',

        # Global options
        'verbose'          : 'Print additional debug messages',
        'quiet'            : 'Suppress all unrequested '
                             'output (including warnings)',
        'logFile'          : 'Write log messages to this file (default: '
                             'log messages are written to standard output).',
        'testDir'          : 'Directory containing tests (default: current '
                             'directory)',
        'includeTests'     : 'Only consider tests matching this wildcard '
                             'pattern. Can be used multiple times. ',
        'excludeTests'     : 'Do not consider tests matching this wildcard '
                             'pattern. Takes precedence over --includeTests. '
                             'Can be used multiple times. ',
        'cacheSize'        : 'Size (MB) to use for the internal image cache '
                             '(default: 32768)',

        # Shared options
        'inputDir'         : 'Location of input data directory',
        'benchmarkDir'     : 'Location of benchmark data directory',
        'outputDir'        : 'Location of output directory',
        'overwrite'        : 'Overwrite output directory if it exists',
        'config'           : 'Pyfeeds configuration file',
        'hashFile'         : 'File to read/write test hashes (default: '
                             '[inputDir]/.feedsHashes',
        'forceHashes'      : 'Force test hashing, even if file modification '
                             'times suggest that the data has not been '
                             'modified since the last check.',

        # Run options
        'jobs'             : 'Number of jobs to run in parallel',

        'updateHashes'     : 'Update test data hashes',
        'skipHashes'       : 'Skip test hash verification',
        'sandboxDir'       : 'Directory to store test sandboxes '
                             '(default: temporary location)',
        'leaveSandboxes'   : 'Do not delete test sandbox directories',
        'selfEval'         : 'List of tests which evaluate themselves. '
                             'Specify the list as a single argument (quoted '
                             'if necessary), with tests separated by the '
                             '{} character.'.format(ARG_SEP),
        'evalRoutines'     : 'List of (file pattern, [evaluation routine]) '
                             'pairs. Specify the list as a single argument '
                             '(quoted if necessary), with pairs separated '
                             'by the {} character, and with each pair '
                             'having the form '
                             '"pattern=routine1,routine2,...".'.format(
                                 ARG_SEP),
        'exclude'          : 'List of file patterns, separated with the {} '
                             'character (and quoted if necessary), '
                             'specifying files which should not be '
                             'evaluated. This list takes precedence over the '
                             'file patterns specified via the "{}" '
                             'argument.'.format(
                                 ARG_SEP, flags['evalRoutines'][1]),
        'tolerances'       : 'List of (file pattern, tolerance) pairs.'
                             'Specify the list as a single argument '
                             '(quoted if necessary), with pairs separated '
                             'by the {} character, and with each pair '
                             'having the form '
                             '"pattern=tolerance".'.format(ARG_SEP),
        'defaultTolerance' : 'Default tolerance to use for files which '
                             'do not have a specific tolerance specified '
                             'via the "{}" argument. '.format(
                                 flags['tolerances'][1]),
        'fileGroups'       : 'List of file patterns denoting groups of '
                             'files which should be evaluated together. '
                             'Specify all groups as a single argument, '
                             'where groups are separated by the {} '
                             'character, and with each group having the '
                             'form "pattern1,pattern2,pattern3,'
                             '...". All files in a group are assumed to '
                             'be in the same directory.'.format(ARG_SEP),

        # Bundle options
        'genHashes'        : 'Generate hashes for bundled tests. The '
                             'hashes are saved in '
                             '[outputDir]/[inputDir]/.feedsHashes',

        # Compare options
        'outputFile'       : 'Save results to this file (default: print to '
                             'standard output).',
    }

    descs = {
        'mainParser' :
        'Pyfeeds allows you to run and manage the FSL test suite. Choose\n'
        'one of the pyfeeds commands, and specify a set of directories\n'
        'which contain tests.',

        'list' :
        'The \'list\' command lists all of the tests that are found in\n'
        'the specified test directories.',

        'run' :
        'The \'run\' command runs and evaluates all detected tests, and\n'
        'prints a summary of passes and failures.',

        'bundle' :
        'The \'bundle\' command copies all detected tests and their\n'
        'associated data into a new stand-alone directory.',

        'genhash' :
        'The \'genhash\' command calculates hashes of all detected tests\n'
        'and test data, for verification purposes.',

        'checkhash' :
        'The \'checkhash\' command calculates hashes of all detected\n'
        'tests and test data, compares them to the stored test hashes,\n'
        'and prints a summary of which tests have changed.',

        'compare' :
        'The \'compare\' command compares two directories to each other,\n'
        'and outputs a summary of the difference between them.',
    }

    epilogs = {
        'mainParser' :
        'To get help on a pyfeeds command, type\n\n'
        '    pyfeeds command --help',

        'list'    : None,
        'bundle'  : None,
        'compare' :
        'Available evaluation routines are:\n{}'.format(
            '\n'.join(['  {}'.format(r) for r in evaluate.listRoutines()])),

        'run' :
        'If the hashFile argument is not provided, the test hashes are \n'
        'assumed to be located at:\n\n'
        '    [inputDir]/.feedsHashes\n\n'
        'Available evaluation routines are:\n{}'.format(
            '\n'.join(['  {}'.format(r) for r in evaluate.listRoutines()])),

        'genhash' :
        'If the hashFile argument is not provided, the test hashes are\n'
        'printed to standard output.',

        'checkhash' :
        'If the hashFile argument is not provided, the test hashes are\n'
        'assumed to be located at:\n\n'
        '    [inputDir]/.feedsHashes',
    }

    mainParser = argparse.ArgumentParser(
        prog='pyfeeds',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=descs['mainParser'],
        usage=usages['mainParser'],
        epilog=epilogs['mainParser'])

    # Global options shared by all sub-commands
    baseParser = argparse.ArgumentParser(add_help=False)
    baseParser.add_argument(*flags['version'],
                            action='version',
                            version=pyfeeds_version)
    baseParser.add_argument(*flags['config'],
                            help=helps['config'])
    baseParser.add_argument(*flags['logFile'],
                            help=helps['logFile'])
    baseParser.add_argument(*flags['includeTests'],
                            help=helps['includeTests'],
                            action='append')
    baseParser.add_argument(*flags['excludeTests'],
                            help=helps['excludeTests'],
                            action='append')
    baseParser.add_argument(*flags['cacheSize'],
                            help=helps['cacheSize'],
                            type=int)

    vqParser = baseParser.add_mutually_exclusive_group()
    vqParser.add_argument(*flags['verbose'],
                          help=helps['verbose'],
                          action='store_true')
    vqParser.add_argument(*flags['quiet'],
                          help=helps['quiet'],
                          action='store_true')

    # Options shared by all sub-commands except 'compare'
    testBaseParser = argparse.ArgumentParser(add_help=False)
    testBaseParser.add_argument(flags['testDir'],
                                help=helps['testDir'],
                                nargs='*')

    # Options shared by the run/bundle/genhash/checkhash sub-commands
    runbungencheckParser = argparse.ArgumentParser(add_help=False)
    runbungencheckParser.add_argument(*flags['inputDir'],
                                      help=helps['inputDir'])

    # Options shared by the run/bundle sub-commands
    runbunParser = argparse.ArgumentParser(add_help=False)
    runbunParser.add_argument(*flags['outputDir'],
                              help=helps['outputDir'])
    runbunParser.add_argument(*flags['overwrite'],
                              help=helps['overwrite'],
                              action='store_true')
    runbunParser.add_argument(*flags['benchmarkDir'],
                              help=helps['benchmarkDir'])

    # Options shared by the run/genhash/checkhash sub-commands
    rungencheckParser = argparse.ArgumentParser(add_help=False)
    rungencheckParser.add_argument(*flags['hashFile'],
                                   help=helps['hashFile'])

    # Options shared by the run/checkhash sub-commands
    runcheckParser = argparse.ArgumentParser(add_help=False)
    runcheckParser.add_argument(*flags['forceHashes'],
                                action='store_true',
                                help=helps['forceHashes'])

    # Options for the compare sub-command
    compareParser = argparse.ArgumentParser(add_help=False)
    compareParser.add_argument('inputDir',
                               help=helps['inputDir'])
    compareParser.add_argument('benchmarkDir',
                               help=helps['benchmarkDir'])
    compareParser.add_argument(*flags['outputFile'],
                               help=helps['outputFile'])

    # Options shared by the run/compare sub-commands
    runcmpParser = argparse.ArgumentParser(add_help=False)
    runcmpParser.add_argument(*flags['selfEval'],
                              help=helps['selfEval'])
    runcmpParser.add_argument(*flags['evalRoutines'],
                              help=helps['evalRoutines'])
    runcmpParser.add_argument(*flags['exclude'],
                              help=helps['exclude'])
    runcmpParser.add_argument(*flags['tolerances'],
                              help=helps['tolerances'])
    runcmpParser.add_argument(*flags['defaultTolerance'],
                              help=helps['defaultTolerance'],
                              type=float)
    runcmpParser.add_argument(*flags['fileGroups'],
                              help=helps['fileGroups'])

    parents = {
        'list'      : [baseParser, testBaseParser],
        'run'       : [baseParser, testBaseParser, runbungencheckParser,
                       runbunParser, rungencheckParser, runcheckParser,
                       runcmpParser],
        'bundle'    : [baseParser, testBaseParser, runbungencheckParser,
                       runbunParser],
        'genhash'   : [baseParser, testBaseParser, runbungencheckParser,
                       rungencheckParser],
        'checkhash' : [baseParser, testBaseParser, runbungencheckParser,
                       rungencheckParser, runcheckParser],
        'compare'   : [baseParser, compareParser, runcmpParser]
    }

    # parsers for each sub-command
    subParserObjs = {}
    subParsers    = mainParser.add_subparsers(title='Pyfeeds commands',
                                              dest='command')

    for cmd in ['list', 'run', 'bundle', 'genhash', 'checkhash', 'compare']:
        subParserObjs[cmd] = subParsers.add_parser(
            cmd,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            parents=parents[cmd],
            usage=usages[cmd],
            help=helps[cmd],
            description=descs[cmd],
            epilog=epilogs[cmd])

    # run options
    runParser = subParserObjs['run']
    runParser.add_argument(*flags['sandboxDir'],
                           help=helps['sandboxDir'])
    runParser.add_argument(*flags['jobs'],
                           help=helps['jobs'],
                           type=int)
    runParser.add_argument(*flags['skipHashes'],
                           help=helps['skipHashes'],
                           action='store_const',
                           const=True)
    runParser.add_argument(*flags['updateHashes'],
                           help=helps['updateHashes'],
                           action='store_const',
                           const=True)
    runParser.add_argument(*flags['leaveSandboxes'],
                           help=helps['leaveSandboxes'],
                           action='store_const',
                           const=True)

    # bundle options
    bundleParser = subParserObjs['bundle']
    bundleParser.add_argument(*flags['genHashes'],
                              help=helps['genHashes'],
                              action='store_const',
                              const=True)

    args = mainParser.parse_args(argv)

    if args.command not in ('run', 'compare'):
        return args

    # The selfEval argument needs to be
    # turned from a string of the form:
    #
    #   test1:test2:test3
    #
    # into a list of test names
    if args.selfEval is not None:
        selfEval      = args.selfEval.split(ARG_SEP)
        selfEval      = [t.strip() for t in selfEval if t is not None]
        args.selfEval = selfEval

    # evalRoutines needs to be turned from
    # a string of the form:
    #
    #   pattern1=routine1:pattern2=routine1,routine2:testpattern3:routine1
    #
    # into a dictionary of
    #
    #   { testName : [routine] }
    #
    # mappings. The routines are kept as
    # strings for the time being - they
    # are validated by the Pyfeeds object.
    if args.evalRoutines is not None:
        evalRoutines = args.evalRoutines.split(ARG_SEP)
        evalRoutines = [r.split('=') for r in evalRoutines]

        for i, (pat, routines) in enumerate(evalRoutines):
            routines        = routines.split(',')
            routines        = [r.strip() for r in routines]
            evalRoutines[i] = (pat, routines)

        args.evalRoutines = OrderedDict(evalRoutines)

    # exclude needs to be turned from a
    # string of the form:
    #
    #   pattern1:pattern2:pattern3
    #
    # into a list of patterns.
    if args.exclude is not None:
        exclude      = args.exclude.split(ARG_SEP)
        exclude      = [p.strip() for p in exclude if p is not None]
        args.exclude = exclude

    # Tolerances needs to be turned from
    # a string of the form:
    #
    #   pattern1=tolerance:pattern2=tolerance:pattern3:tolerance
    #
    # into a dictionary of
    #
    #   { patern : tolerance }
    #
    # mappings.
    if args.tolerances is not None:
        tolerances      = args.tolerances.split(ARG_SEP)
        tolerances      = [t.split('=') for t in tolerances]
        tolerances      = [(n, float(t)) for (n, t) in tolerances]
        args.tolerances = OrderedDict(tolerances)

    # fileGroups needs to be turned from a
    # string of the form:
    #
    #   fg1pat1,fg1pat2:fg2pat1,fg2pat2,fg2pat3
    #
    # into a list of file pattern groups.
    if args.fileGroups is not None:
        fileGroups      = args.fileGroups.split(ARG_SEP)
        fileGroups      = [fg.split(',') for fg in fileGroups]
        fileGroups      = [[p.strip() for p in fg] for fg in fileGroups]
        args.fileGroups = fileGroups

    return args


def loadPyfeedsConfig(cfgfile):
    """Loads settings from the given pyfeeds configuration file.

    :arg cfgfile: Path to a pyfeeds configuration file.

    :returns:     An ``argparse.Namespace`` object containing all of the
                  settings that were present in the file
    """

    args = argparse.Namespace()

    if cfgfile is None:
        return args

    if not op.exists(cfgfile):
        raise RuntimeError('{} does not exist!'.format(cfgfile))

    def strbool(s):
        if s.lower() == 'false': return False
        else:                    return bool(s)

    def strlist(s):
        items = [i.strip() for i in s.split('\n')]
        items = [i for i in items if i != '']
        return items

    def strlistlist(s):
        items = [i.strip()                 for i in s.split('\n')]
        items = [i                         for i in items if i != '']
        items = [i.split(',')              for i in items]
        items = [[ii.strip()  for ii in i] for i in items]
        return items

    types = {
        'verbose'          : strbool,
        'quiet'            : strbool,
        'updateHashes'     : strbool,
        'skipHashes'       : strbool,
        'forceHashes'      : strbool,
        'leaveSandboxes'   : strbool,
        'genHashes'        : strbool,
        'defaultTolerance' : float,
        'jobs'             : int,
        'selfEval'         : strlist,
        'testDir'          : strlist,
        'exclude'          : strlist,
        'fileGroups'       : strlistlist,
        'includeTests'     : strlist,
        'excludeTests'     : strlist
    }

    # Use a case-sensitive file parser
    cfg             = configparser.ConfigParser()
    cfg.optionxform = str
    cfg.read(cfgfile)

    generalSettings     = cfg.options('general')
    evalRoutinePatterns = cfg.options('evalRoutines')
    tolerancePatterns   = cfg.options('tolerances')

    evalRoutines = [cfg.get('evalRoutines', p)
                    for p in evalRoutinePatterns]

    evalRoutines = [r.split(',') for r in evalRoutines]
    evalRoutines = [[r.strip() for r in er] for er in evalRoutines]
    evalRoutines = reversed(list(zip(evalRoutinePatterns, evalRoutines)))
    evalRoutines = OrderedDict(evalRoutines)

    tolerances   = [cfg.getfloat('tolerances', p)
                    for p in tolerancePatterns]
    tolerances   = reversed(list(zip(tolerancePatterns,   tolerances)))
    tolerances   = OrderedDict(tolerances)

    if len(evalRoutines) > 0: args.evalRoutines = evalRoutines
    if len(tolerances)   > 0: args.tolerances   = tolerances

    for key in generalSettings:

        typecnv = types.get(key, str)
        val     = typecnv(cfg.get('general', key))

        setattr(args, key, val)

    return args


if __name__ == '__main__':
    """Calls the :func:`main` function. """
    main()



#!/usr/bin/env python
#
# hashing.py - Pyfeeds test verification
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions used by pyfeeds to perform **test
verificiation** - identifying when a test or its data has changed since it was
last executed.


Pyfeeds implements a simple change-tracking mechanism so it can warn the user
when the contents of a test directory, or the external data dependencies of
a test have changed. The following functions are provided to use this
mechanism:

 .. autosummary::
    :nosignatures:

    genHashes
    checkHashes
    calcHashes
    loadHashes
    saveHashes
    checkTestHash
    hashPath
    hashFile
"""


from __future__ import print_function
from __future__ import division

import            os
import os.path as op
import            logging
import            hashlib


from . import common


log = logging.getLogger(__name__)


def genHashes(pyf):
    """Generates verification hashes for all of the given tests, and
    prints them to standard output, or writes them to a file.

    :arg pyf: A :class:`.Pyfeeds` instance.
    """

    tests    = pyf.tests
    inputDir = pyf.inputDir

    testHashes  = []
    inputHashes = []

    log.info('Calculating test hashes ...')

    testHashes, inputHashes = calcHashes(tests, inputDir)

    # Hashes can be None, so make
    # sure they're formatted as strings
    testHashes  = [str(h) for h in testHashes]
    inputHashes = [str(h) for h in inputHashes]

    # Write hashes to hashFile
    if pyf.hashFileProvided:

        log.info('Saving test hashes to {} ...'.format(pyf.hashFile))

        hashes = {th : ih for th, ih in zip(testHashes, inputHashes)}
        saveHashes(pyf.hashFile, hashes)

    # Otherwise, print hashes to standard output
    else:
        testNames  = [t.name for t in tests]
        titles     = ['Test name', 'Test directory hash', 'Test data hash']

        print('\nTest verification hashes\n')
        common.printColumns(titles, [testNames, testHashes, inputHashes])


def checkHashes(pyf):
    """Calculates verification hashes for all of the given tests, compares
    them against the saved hashes, then prints a summary to standard output
    outlining which tests have changed.

    :arg pyf: A :class:`.Pyfeeds` instance.
    """

    tests    = pyf.tests
    inputDir = pyf.inputDir
    force    = pyf.forceHashes
    hashFile = pyf.hashFile

    if hashFile is not None:
        hashFile = op.abspath(hashFile)
    savedHashes = loadHashes(hashFile)

    results = [checkTestHash(pyf, t, inputDir, savedHashes, False, force)
               for t in tests]
    results = [TEST_HASH_STRINGS[r] for r in results]
    names   = [t.name for t in tests]

    titles = ['Test name', 'Verification test result']

    print('\nTest verification results\n')
    common.printColumns(titles, [names, results])


def calcHashes(tests, inputDir):
    """Calculates verification hashes for all of the given tests. A tuple
    containing the following is returned:

      - A list of the test directory hashes
      - A list of the test input hashes

    :arg tests:    A list of :class:`.Test` object.
    :arg inputDir: Directory in which the shared data is located.
    """

    if inputDir is not None:
        inputDir = op.abspath(inputDir)

    testHashes  = []
    inputHashes = []

    for test in tests:

        log.debug('Calculating hashes for test {} ...'.format(test.name))

        testHashes.append(test.hashTestDir())

        if inputDir is None: inputHashes.append(None)
        else:                inputHashes.append(test.hashInputs(inputDir))

    return testHashes, inputHashes


def loadHashes(hashFile):
    """Loads verification hashes from the given file. The hashes are returned
    as a dictionary of ``{testHash : dataHash}`` mappings.

    :arg hashFile: File to load the hashes from (e.g. ``.feedsHashes``).
    """

    if hashFile is None or not op.exists(hashFile):
        return {}

    log.debug('Loading test data hashes from {}'.format(hashFile))

    hashes = {}

    with open(hashFile, 'rt') as f:

        for i, line in enumerate(f):

            if line.strip() == '':
                continue

            try:
                testHash, inputHash = line.split()

                # 'None' in the hash file indicates
                # that the test does not have any
                # shared data
                if inputHash == 'None':
                    inputHash = None

            except Exception:
                log.warning('Malformed line in {} [{}]: {}'.format(
                    hashFile, i, line))
                continue

            hashes[testHash]  = inputHash

    return hashes


def saveHashes(hashFile, hashes):
    """Saves the given verification hashes to the specified file.

    :arg hashFile: File to save the hashes to (e.g. ``.feedsHashes``).

    :arg hashes:   A dictionary of verification hashes, in the same format
                   as that returned by the :func:`loadHashes` function.
    """

    log.debug('Saving test data hashes to {}'.format(hashFile))

    lines = ['{} {}'.format(th, ih) for th, ih in hashes.items()]

    with open(hashFile, 'wt') as f:

        f.write('\n'.join(lines) + '\n')


TEST_HASH_NO_INPUTS = 0
"""Return code for the :func:`checkTestHash` function. This code indicates
that the test hash was valid, and that the test does not have any external
data dependencies, hence has no data hash.
"""


TEST_HASH_PASSED = 1
"""Return code for the :func:`checkTestHash` function. This code indicates
that both the test hash and the data hash were valid.
"""


TEST_HASH_SKIPPED_NOCHANGE = 2
"""Return code for the :func:`checkTestHash` function. This code indicates
that the test verification was skipped , as the test input data had not been
modified since the last time the hashes were generated.
"""


TEST_HASH_FAILED_TEST_HASH = 3
"""Return code for the :func:`checkTestHash` function. This code indicates
that the test hash could not be found in the dictionary, either because
this is a new test, or because the contents of the test directory have
changed.
"""


TEST_HASH_FAILED_DATA_HASH = 4
"""Return code for the :func:`checkTestHash` function. This code indicates
that the calculated data hash does not match the stored hash, meaning
that the data for this test has changed since the hashes were last calculated.
"""


TEST_HASH_NOT_CHECKED = -1
"""This is a dummy code, used by the :func:`runTests` function to indicate
that the data integrity check has been skipped.
"""


TEST_HASH_STRINGS = {
    TEST_HASH_NOT_CHECKED       : 'Not checked',
    TEST_HASH_PASSED            : 'Passed',
    TEST_HASH_SKIPPED_NOCHANGE  : 'Skipped (data not changed since last '
                                  'check)',
    TEST_HASH_NO_INPUTS         : 'Passed (test does not use any shared data)',
    TEST_HASH_FAILED_TEST_HASH  : 'Failed (test directory changed, or this is '
                                  'a new test)',
    TEST_HASH_FAILED_DATA_HASH  : 'Failed (shared test data changed)',
}
"""Human-readable descriptions for each of the test hash results. """


TEST_HASH_SHORT_STRINGS = {
    TEST_HASH_NOT_CHECKED       : 'Not checked',
    TEST_HASH_PASSED            : 'Passed',
    TEST_HASH_SKIPPED_NOCHANGE  : 'Skipped (no change)',
    TEST_HASH_NO_INPUTS         : 'No data',
    TEST_HASH_FAILED_TEST_HASH  : 'New test, or test has changed',
    TEST_HASH_FAILED_DATA_HASH  : 'Data has changed',
}
"""Shorter human-readable descriptions for the test hash results. """


def checkTestHash(pyf,
                  test,
                  inputDir,
                  testHashes,
                  updateHashes,
                  forceCheck=False):
    """Calculates verification hashes for the given :class:`.Test`, and compares
    them against the hashes stored in the ``testHashes`` dictionary.  One of
    the following codes is returned, indicating the outcome of the
    verification check:

    .. autosummary::

       TEST_HASH_NO_INPUTS
       TEST_HASH_PASSED
       TEST_HASH_SKIPPED_NOCHANGE
       TEST_HASH_FAILED_TEST_HASH
       TEST_HASH_FAILED_DATA_HASH

    :arg pyf:         The :class:`.Pyfeeds` instance.

    :arg test:         The :class:`.Test` object to check.

    :arg inputDir:     Directory in which the shared data is located.

    :arg testHashes:   Dictionary of stored test hashes, in the format returned
                       by the :func:`loadHashes` function.

    :arg updateHashes: If ``True``, and the verification fails, the
                       ``testHashes`` dictionary is updated with the newly
                       calculated test/data hash values.

    :arg forceCheck:   If ``True``, the check is always performed, even if the
                       test data file timestamps indicate that they have not
                       not changed.
    """

    log.debug('Verifying input data for test {}'.format(test))

    # Generate hashes for the test directory, and
    # retrieve the stored input data hash from the
    # testHashes dictionary. If the testHash is
    # not in the dictionary, oldInputHash will be
    # a reference to the test object. We do this
    # because 'None' is a valid value for a data
    # hash.
    testHash     = test.hashTestDir()
    oldInputHash = testHashes.get(testHash, test)

    # If a hash has previously been calculated
    # on this test, check to see if the test
    # inputs have changed at all since the last
    # hash calculation time. If they haven't
    # we're going to skip the input hash
    # calculation.
    if (inputDir is not None) and \
       (not forceCheck)       and \
       (oldInputHash is not test):
        if not test.inputsChangedSince(inputDir, pyf.hashesLastGenerated):
            log.debug('Test %s data has not changed since '
                      'the last hash calculation - skipping.', test)
            return TEST_HASH_SKIPPED_NOCHANGE

    # Generate hashes for the test data
    if inputDir is not None:
        newInputHash = test.hashInputs(inputDir)
    else:
        newInputHash = None

    log.debug('Test {} input check: testHash={}, '
              'oldInputHash={}, newInputHash={}'.format(
                  test.name, testHash, newInputHash, oldInputHash))

    # This test has no inputs, and hence no hash
    if (newInputHash is None) and (oldInputHash is None):
        return TEST_HASH_NO_INPUTS

    # The hashes match
    if newInputHash == oldInputHash:
        return TEST_HASH_PASSED

    # The test directory hash could not
    # be found in the testHashes dict
    result = None
    if oldInputHash is test:

        result = TEST_HASH_FAILED_TEST_HASH

        log.warning('Test {} hashes not found in dictionary! This either '
                    'means that this is a new test which has not previously '
                    'been verified (in which case this warning can be '
                    'ignored), or that the contents of the test directory has '
                    'changed since it was last verified.'.format(test.name))

    # The test input data
    # hashes do not match
    else:
        result = TEST_HASH_FAILED_DATA_HASH

        log.warning('Test {} data verification failed! It looks like the '
                    'shared data used by this test has changed since the '
                    'test was last run.'.format(test.name))

    if updateHashes:
        log.debug('The updateHashes flag is set - adding/overwriting '
                  'the hash values for test {}.'.format(test.name))
        testHashes[testHash] = newInputHash

    return result


def hashPath(path, hashObj):
    """Recursively calculates MD5 digests of every file found from the
    given starting point. If ``path`` is a file, the ``hashObj`` is
    updated with the hex digest of its contents - a hash of hashes (see
    the :func:`hashFile` function). Otherwise, if ``path`` is a directory,
    this method is recursively called with its children.

    :arg path:    A file or directory path.

    :arg hashObj: A ``hashlib.md5`` object.

    :returns:     The number of files which were hashed.
    """

    if op.isfile(path):
        count = 1
        hashObj.update(hashFile(path).encode('ascii'))

    elif op.isdir(path):

        # listdir does not guarantee a specific
        # ordering, so we sort the result to
        # ensure a consistent hash value
        count = 0
        for p in sorted(os.listdir(path)):

            p      = op.join(path, p)
            count += hashPath(p, hashObj)

    else:

        count = 0
        log.warning('Attempt to hash invalid path: "{}"'.format(path))

    return count



hashCache = {}
"""This dictionary is used by the :meth:`hashFile` method. Whenever a md5
digest is calculated on a file, it is stored in this dictionary. Before a
file is hashed, this cache is checked, and the cached value (if present)
is used instead of re-calculating the hash. The dictionary structure is
``{ filePath : hexdigest }``.
"""


def hashFile(path):
    """Calculates an MD5 digest of the given file path.

    :arg path: Path to a file to be hashed.

    :returns:  A string containing the hexadecimal digest.
    """

    cached = hashCache.get(path)

    if cached is not None:
        log.debug('Using cached md5 digest for {}'.format(path))
        return cached

    hashObj = hashlib.md5()

    log.debug('Calculating md5 digest for {}'.format(path))

    with open(path, 'rb') as f:
        hashObj.update(f.read())

    digest          = hashObj.hexdigest()
    hashCache[path] = digest

    return digest



#!/usr/bin/env python
#
# test_testing.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os
import datetime
import os.path as op

from . import tempdir, makepaths, maketest, makepyfeeds, CaptureStdout

from  pyfeeds import testing


def test_findTestDirs():
    with tempdir():
        makepaths(('test1/feedsRun',
                   'test2/feedsRun',
                   'test3/a/feedsRun',
                   'test3/b/feedsRun',
                   'test4/feedsRun',
                   'test4/feedsRun.a',
                   'test4/feedsRun.b'))
        got = list(sorted(testing.findTestDirs(os.getcwd())))
        exp = ['test1',
               'test2',
               'test3/a',
               'test3/b',
               'test4']

        got = [op.abspath(p) for p in got]
        exp = [op.abspath(p) for p in exp]

        assert got == exp


def test_findTestScripts():
    def check(got, exp):
        got = list(sorted([op.abspath(p) for p in got]))
        exp = list(sorted([op.abspath(p) for p in exp]))
        assert got == exp

    tests = [
        ['./feedsRun'],
        ['./feedsRun.a', './feedsRun..b'],
        ['./feedsRun', './feedsRun.a', './feedsRun.b']]


    for paths in tests:
        with tempdir():
            makepaths(paths)
            got = testing.findTestScripts('.')
            check(got, paths)

    with tempdir():
        makepaths(['feedsRu', 'feedsRun.', '.feedsRun'])
        assert testing.findTestScripts('.') == []



def test_listTests():
    cap = CaptureStdout()
    with tempdir(), cap:

        tests = ['test1/feedsRun',
                 'test2/feedsRun',
                 'test3/subdir1/feedsRun',
                 'test3/subdir2/feedsRun',
                 'test4/feedsRun',
                 'test4/feedsRun.abc',
                 'test4/feedsRun.def']

        makepaths(tests)

        pyf = makepyfeeds(command='list', testDir=['.'])
        testing.listTests(pyf)

        tkns = cap.stdout.split()

        for test in tests:
            dirname = op.dirname(test)
            assert dirname             in tkns
            assert op.abspath(dirname) in tkns



def test_Test():
    with tempdir():
        maketest('test1/feedsRun', inputs=['data/file1', 'data/file2'])

        test = testing.Test('test1/feedsRun', '.')

        assert test.name  == 'test1'
        assert str(test)  == 'test1'
        assert repr(test) == 'test1'
        assert test.getInputs() == ['data/file1', 'data/file2']

        makepaths(['inputData/data/file1', 'inputData/data/file2'])
        test.inputsChangedSince('inputData', datetime.datetime.now())
        test.hashTestDir()
        test.hashInputs('inputData')
        os.mkdir('sandbox')
        test.createSandbox('inputData', 'sandbox')

        maketest('test1/subdir/feedsRun')
        test = testing.Test('test1/subdir/feedsRun', '.')
        assert test.name == 'test1/subdir'

        maketest('test1/subdir/feedsRun')
        maketest('test1/subdir/feedsRun.abc')
        maketest('test1/subdir/feedsRun.def')
        test = testing.Test('test1/subdir/feedsRun', '.')
        assert test.name == 'test1/subdir'
        test = testing.Test('test1/subdir/feedsRun.abc', '.')
        assert test.name == 'test1/subdir/abc'
        test = testing.Test('test1/subdir/feedsRun.def', '.')
        assert test.name == 'test1/subdir/def'


def test_bundleTests():
    with tempdir():
        maketest(  'tests/test1/feedsRun',
                   inputs=['data/file1', 'data/file2'])
        maketest(  'tests/test2/feedsRun.1')
        maketest(  'tests/test2/feedsRun.2')

        makepaths(['tests/test1/datafile',
                   'inputData/data/file1',
                   'inputData/data/file2',
                   'benchmarkData/test1/outfile1',
                   'benchmarkData/test1/outfile2',
                   'benchmarkData/test2/1/outfile1',
                   'benchmarkData/test2/2/outfile2'])

        pyf = makepyfeeds(command='bundle',
                          testDir=['tests'],
                          inputDir='inputData',
                          benchmarkDir='benchmarkData',
                          outputDir='bundleDir',
                          genHashes=True,
                          overwrite=True)

        testing.bundleTests(pyf)

        assert op.exists('bundleDir/tests/test1/feedsRun')
        assert op.exists('bundleDir/tests/test1/datafile')
        assert op.exists('bundleDir/data/.feedsHashes')
        assert op.exists('bundleDir/data/data/file1')
        assert op.exists('bundleDir/data/data/file2')
        assert op.exists('bundleDir/benchmarks/test1/outfile1')
        assert op.exists('bundleDir/benchmarks/test1/outfile2')
        assert op.exists('bundleDir/benchmarks/test2/1/outfile1')
        assert op.exists('bundleDir/benchmarks/test2/2/outfile2')

        pyf = makepyfeeds(command='bundle',
                          testDir=['tests'],
                          benchmarkDir='benchmarkData',
                          outputDir='bundleDir',
                          overwrite=True)
        testing.bundleTests(pyf)
        assert     op.exists('bundleDir/tests/test1/feedsRun')
        assert     op.exists('bundleDir/tests/test1/datafile')
        assert not op.exists('bundleDir/data')
        assert     op.exists('bundleDir/benchmarks/test1/outfile1')
        assert     op.exists('bundleDir/benchmarks/test1/outfile2')
        assert     op.exists('bundleDir/benchmarks/test2/1/outfile1')
        assert     op.exists('bundleDir/benchmarks/test2/2/outfile2')

        pyf = makepyfeeds(command='bundle',
                          testDir=['tests'],
                          outputDir='bundleDir',
                          overwrite=True)
        testing.bundleTests(pyf)
        assert     op.exists('bundleDir/tests/test1/feedsRun')
        assert     op.exists('bundleDir/tests/test1/datafile')
        assert not op.exists('bundleDir/data')
        assert not op.exists('bundleDir/benchmarks')


def test_runTest():
    with tempdir():
        maketest('tests/test1/feedsRun', returnCode=0, stdout=['output1', 'output2'])
        maketest('tests/test2/feedsRun', returnCode=1)
        os.mkdir('output/')
        os.chmod('tests/test1/feedsRun', 0o755)
        os.chmod('tests/test2/feedsRun', 0o755)

        pyf = makepyfeeds(command='run',
                          testDir=['tests'],
                          outputDir='output',
                          overwrite=True)

        gott1 = testing.runTest(pyf,
                                pyf.tests[0],
                                'output',
                                None,
                                None,
                                os.environ.copy(),
                                False)
        gott2 = testing.runTest(pyf,
                                pyf.tests[1],
                                'output',
                                None,
                                None,
                                os.environ.copy(),
                                False)

        assert     gott1[0]
        assert not gott2[0]

        with open('output/test1/feedsRun.log') as f:
            assert f.read().strip() == 'output1\noutput2'


def test_runTest_suffix():
    with tempdir():
        maketest('tests/test1/feedsRun.suff1', stdout=['output1'], outputs=['out1'])
        maketest('tests/test1/feedsRun.suff2', stdout=['output2'], outputs=['out2'])
        os.chmod('tests/test1/feedsRun.suff1', 0o755)
        os.chmod('tests/test1/feedsRun.suff2', 0o755)
        os.mkdir('output/')

        pyf = makepyfeeds(command='run',
                          testDir=['tests'],
                          outputDir='output',
                          overwrite=True)

        got = []
        for test in pyf.tests:
            got.append(testing.runTest(pyf,
                                       test,
                                       'output',
                                       None,
                                       None,
                                       os.environ.copy(),
                                       False))

        assert got == [(True, 0), (True, 0)]

        assert op.exists('output/test1/suff1/out1')
        assert op.exists('output/test1/suff2/out2')
        with open('output/test1/suff1/feedsRun.log') as f:
            assert f.read().strip() == 'output1'
        with open('output/test1/suff2/feedsRun.log') as f:
            assert f.read().strip() == 'output2'


def test_runTests():
    cap = CaptureStdout()
    with tempdir(), cap:
        maketest('tests/test1/feedsRun', returnCode=0, stdout=['output1', 'output2'])
        maketest('tests/test2/feedsRun', returnCode=1)
        os.mkdir('output/')
        os.chmod('tests/test1/feedsRun', 0o755)
        os.chmod('tests/test2/feedsRun', 0o755)

        pyf = makepyfeeds(command='run',
                          testDir=['tests'],
                          outputDir='output',
                          overwrite=True)

        testing.runTests(pyf)

    lines = cap.stdout.split('\n')
    assert 'test1' in lines[2] and 'Passed' in lines[2]
    assert 'test2' in lines[3] and 'Failed' in lines[3]



#!/usr/bin/env python
#
# setup.py - setuptools configuration for installing pyfeeds.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os.path as op

from setuptools import setup
from setuptools import find_packages


basedir = op.dirname(__file__)

with open(op.join(basedir, 'requirements.txt'), 'rt') as f:
    install_requires = [i.strip() for i in f.readlines()]

with open(op.join(basedir, 'README.md'), 'rt') as f:
    readme = f.read()

version = {}
with open(op.join(basedir, "pyfeeds", "__init__.py")) as f:
    for line in f:
        if line.startswith('__version__'):
            exec(line, version)
            break
version = version['__version__']

setup(

    name='fsl-pyfeeds',
    version=version,
    description='FSL testing framework',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://git.fmrib.ox.ac.uk/fsl/pyfeeds',
    author='Paul McCarthy',
    author_email='pauldmccarthy@gmail.com',
    license='Apache License Version 2.0',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: Free for non-commercial use',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules'],

    packages=find_packages(exclude=('doc')),
    install_requires=install_requires,
    entry_points={
        'console_scripts' : [
            'pyfeeds = pyfeeds.main:main',
        ]
    }
)
